#
# chipdlg.tcl - edit accessibility and z-coordinate of chip
# 
# Copyright (C) 2005  Toni Ronkko
# 
# Peach Gardens is free software; you can redistribute it and/or modify 
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Peach Gardens is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# Dec 30, 2004  Toni Ronkko <toni at users.playstation2-linux.com>
# 
# $Id: chipdlg.tcl,v 1.2 2005/01/16 23:33:21 tronkko Exp $
#
# $Log: chipdlg.tcl,v $
# Revision 1.2  2005/01/16 23:33:21  tronkko
# select background chip by mouse, more checks on background chip,
# compute suitable zoom factor automatically
#
# Revision 1.1  2005/01/09 20:51:39  tronkko
# Initial revision
#
#
#

# chip zoom factor in chip edit dialog, changed dynamically
set chipzoom 8

# image for the z-coordinate window
set chip_z_img ""

# currently selected depth coordinate
set chip_curz 0

# name of entry containing depth coordinate
set chip_depth_widget ""

# name of entry containing background chip
set chip_bg_widget ""

# images of the accessibility page
for {set i 1} {$i < 10} {incr i} {
  set chip_access_img($i) ""
}

# widget name of chip access list
set chip_access_menu ""

# temporary accessibility
for {set i 1} {$i < 10} {incr i} {
  set chip_access($i) "G"
}


#
# Open dialog for editing chip properties.
#
proc edit_chip { id } {
  global chipw
  global chiph
  global chipzoom
  global chip_image
  global chip_z_img
  global chip_curz
  global chip_depth_widget
  global chip_bg_widget
  global chip_access_img
  global chip_access
  global chip_depth
  global chip_bg
  global chip_terrain

  set win .editchip

  # User tries to enter dialog when existing dialog is open, set
  # background chip.
  if { [winfo exists $win] } {
    status "Using chip #$id as a background chip"
    $chip_bg_widget delete 0 end
    $chip_bg_widget insert 0 $id
    focus $chip_bg_widget
    tabnotebook_display $win.nb "Transparency"
    return
  }  

  # fetch depth coordinate of the chip
  set chip_curz $chip_depth($id)

  # get id of background chip
  set bg $chip_bg($id)

  # get chip accessibility
  for {set i 1} {$i < 10} {incr i} {
    set chip_access($i) $chip_terrain($id,$i)
  }

  # determinate suitable zoom value
  if { $chipw > $chiph } {
    set chipzoom [expr int(250/$chipw)]
  } else {
    set chipzoom [expr int(250/$chiph)]
  }

  
  ### LAY COMPONENTS ###
  
  # create toplevel window for the dialog
  catch {destroy $win}
  toplevel $win
  wm title $win "Edit Chip..."
  wm iconname $win "Edit Chip..."

  # prevent dialog window from being resized
  wm resizable $win 0 0 

  # tabbed notebook
  tabnotebook_create $win.nb
  set page1 [tabnotebook_page $win.nb "Accessibility"]
  set page2 [tabnotebook_page $win.nb "Transparency"]

  # divide chip into 9 fragments  
  set w1 [expr int($chipw/3)*$chipzoom]
  set w2 [expr $chipw*$chipzoom - 2*$w1]
  set w3 $w1
  set h1 [expr int($chiph/3)*$chipzoom]
  set h2 [expr $chiph*$chipzoom - 2*$h1]
  set h3 $h1
  
  # accessibility
  frame $page1.access
  canvas $page1.access.b1 -relief raised -bd 2 -width $w1 -height $h1 \
      -takefocus 1
  canvas $page1.access.b2 -relief raised -bd 2 -width $w2 -height $h1 \
      -takefocus 1
  canvas $page1.access.b3 -relief raised -bd 2 -width $w3 -height $h1 \
      -takefocus 1
  canvas $page1.access.b4 -relief raised -bd 2 -width $w1 -height $h2 \
      -takefocus 1
  canvas $page1.access.b5 -relief raised -bd 2 -width $w2 -height $h2 \
      -takefocus 1
  canvas $page1.access.b6 -relief raised -bd 2 -width $w3 -height $h2 \
      -takefocus 1
  canvas $page1.access.b7 -relief raised -bd 2 -width $w1 -height $h3 \
      -takefocus 1
  canvas $page1.access.b8 -relief raised -bd 2 -width $w2 -height $h3 \
      -takefocus 1
  canvas $page1.access.b9 -relief raised -bd 2 -width $w3 -height $h3 \
      -takefocus 1

  # transparency
  frame $page2.top
  set c $page2.top.canvas
  canvas $c \
      -width [expr $chipw*$chipzoom] \
      -height [expr $chiph*$chipzoom] \
      -relief sunken -bd 2
  frame $page2.bottom  
  label $page2.bottom.lab1 -text "Depth:"
  set chip_depth_widget $page2.bottom.z
  entry $chip_depth_widget -width 4
  label $page2.bottom.lab2 -text "Background:"
  set chip_bg_widget $page2.bottom.bg
  entry $chip_bg_widget -width 4

  # separator
  frame $win.separ -relief sunken -bd 2 -height 2
  
  # button row
  frame $win.buttons
  button $win.buttons.ok -text "OK" -command "chip_edit_ok $win $id"
  button $win.buttons.cancel -text "Cancel" -command "destroy $win"
  button $win.buttons.help -text "Help" -command "chip_help $win"
  

  ### PACK COMPONENTS ###
  
  # pack page 1 (access)
  pack $page1.access -side top -padx 8 -pady 12
  grid $page1.access.b1 -row 0 -column 0 -padx 2 -pady 2
  grid $page1.access.b2 -row 0 -column 1 -padx 2 -pady 2
  grid $page1.access.b3 -row 0 -column 2 -padx 2 -pady 2
  grid $page1.access.b4 -row 1 -column 0 -padx 2 -pady 2
  grid $page1.access.b5 -row 1 -column 1 -padx 2 -pady 2
  grid $page1.access.b6 -row 1 -column 2 -padx 2 -pady 2
  grid $page1.access.b7 -row 2 -column 0 -padx 2 -pady 2
  grid $page1.access.b8 -row 2 -column 1 -padx 2 -pady 2
  grid $page1.access.b9 -row 2 -column 2 -padx 2 -pady 2

  # pack page 2 (transparency)
  pack $page2.top -padx 4 -pady 4 -expand yes -fill both -anchor ne
  pack $c
  pack $page2.bottom -fill x -expand yes -padx 4 -pady 4 -anchor se
  grid $page2.bottom.lab1 -column 0 -row 0 -sticky nw
  grid $chip_depth_widget -column 1 -row 0 -sticky nwe
  grid $page2.bottom.lab2 -column 0 -row 1 -sticky nw
  grid $chip_bg_widget -column 1 -row 1 -sticky nwe
  grid columnconfigure $page2.bottom 1 -weight 1

  # pack buttons
  pack $win.buttons.ok -side left -padx 4
  pack $win.buttons.cancel -side left -padx 4
  pack $win.buttons.help -side left -padx 4
  
  # pack major components
  pack $win.nb -expand yes -fill both
  pack $win.separ -fill x -padx 4 -pady 4
  pack $win.buttons -side bottom -padx 4 -pady 4


  ### FILL ENTRIES ###
  
  # compute fragment coordinates for splicing input image to access window
  set x1 [expr int($w1/$chipzoom)]
  set x2 [expr $x1 + int($w2/$chipzoom)]
  set x3 [expr $x2 + int($w3/$chipzoom)]
  set y1 [expr int($h1/$chipzoom)]
  set y2 [expr $y1 + int($h2/$chipzoom)]
  set y3 [expr $y2 + int($h3/$chipzoom)]

  # create images in accessibility buttons
  for {set i 1} {$i < 10} {incr i} {
    if { $chip_access_img($i) == "" } {
      set chip_access_img($i) [image create photo]
    }
    $chip_access_img(1) blank
  }
  $chip_access_img(1) copy $chip_image($id,0) -zoom $chipzoom $chipzoom \
      -from 0 0 $x1 $y1
  $chip_access_img(2) copy $chip_image($id,0) -zoom $chipzoom $chipzoom \
      -from $x1 0 $x2 $y1
  $chip_access_img(3) copy $chip_image($id,0) -zoom $chipzoom $chipzoom \
      -from $x2 0 $x3 $y1
  $chip_access_img(4) copy $chip_image($id,0) -zoom $chipzoom $chipzoom \
      -from 0 $y1 $x1 $y2
  $chip_access_img(5) copy $chip_image($id,0) -zoom $chipzoom $chipzoom \
      -from $x1 $y1 $x2 $y2
  $chip_access_img(6) copy $chip_image($id,0) -zoom $chipzoom $chipzoom \
      -from $x2 $y1 $x3 $y2
  $chip_access_img(7) copy $chip_image($id,0) -zoom $chipzoom $chipzoom \
      -from 0 $y2 $x1 $y3
  $chip_access_img(8) copy $chip_image($id,0) -zoom $chipzoom $chipzoom \
      -from $x1 $y2 $x2 $y3
  $chip_access_img(9) copy $chip_image($id,0) -zoom $chipzoom $chipzoom \
      -from $x2 $y2 $x3 $y3
  
  # set images of accessibility buttons
  for {set i 1} {$i < 10} {incr i} {    
    $page1.access.b$i create image 0 0 -anchor nw -image $chip_access_img($i)
    $page1.access.b$i xview moveto 0
    $page1.access.b$i yview moveto 0
    chip_show_access $page1.access.b$i $i
  }

  # insert image to transparency window
  if { $chip_z_img == "" } {
    set chip_z_img [image create photo]
  }
  $chip_z_img blank
  $chip_z_img copy $chip_image($id,0) -zoom $chipzoom $chipzoom
  $c create image 0 0 -anchor nw -image $chip_z_img
  $c xview moveto 0
  $c yview moveto 0

  # paint accessibility limits to transparency window
  set height [expr $chiph*$chipzoom]
  set width  [expr $chipw*$chipzoom]
  $c create line \
      0 [expr $y1*$chipzoom] $width [expr $y1*$chipzoom] \
      -stipple gray50 -fill white
  $c create line \
      0 [expr $y2*$chipzoom] $width [expr $y2*$chipzoom] \
      -stipple gray50 -fill white 
  $c create line \
      [expr $x1*$chipzoom] 0 [expr $x1*$chipzoom] $height \
      -stipple gray50 -fill white
  $c create line \
      [expr $x2*$chipzoom] 0 [expr $x2*$chipzoom] $height \
      -stipple gray50 -fill white 
  
  # show current z-coordinate in transparency window
  show_z $c

  # fill entries
  $chip_depth_widget insert 0 $chip_curz
  $chip_bg_widget    insert 0 $bg


  ### BIND EVENTS ###

  # bind accessibility buttons
  for {set i 1} {$i < 10} {incr i} {    
    bind $page1.access.b$i <2> \
        "chip_rotate_access $page1.access.b$i $i"
    bind $page1.access.b$i <1> \
        "chip_swap_access $page1.access.b$i $i"
    bind $page1.access.b$i <3> \
        "chip_change_access $page1.access.b$i $i"
    bind $page1.access.b$i <KeyPress-space> \
        "chip_change_access $page1.access.b$i $i"
    bind $page1.access.b$i <KeyPress-Return> \
        "chip_change_access $page1.access.b$i $i"
    bind $page1.access.b$i <Enter> \
        "access_hint $page1.access.b$i $i"
    bind $page1.access.b$i <Leave> "status {}"
    bind $page1.access.b$i <KeyPress-x> \
        "chip_set_access $page1.access.b$i $i U"
    bind $page1.access.b$i <KeyPress-u> \
        "chip_set_access $page1.access.b$i $i U"
    bind $page1.access.b$i <KeyPress-g> \
        "chip_set_access $page1.access.b$i $i G"
    bind $page1.access.b$i <KeyPress-w> \
        "chip_set_access $page1.access.b$i $i W"
    bind $page1.access.b$i <KeyPress-f> \
        "chip_set_access $page1.access.b$i $i F"
    bind $page1.access.b$i <KeyPress-m> \
        "chip_set_access $page1.access.b$i $i M"
    bind $page1.access.b$i <KeyPress-r> \
        "chip_set_access $page1.access.b$i $i R"
  }

  # keyboard traversal
  bind $page1.access.b1 <Right> "focus $page1.access.b2"
  bind $page1.access.b2 <Right> "focus $page1.access.b3"
  bind $page1.access.b4 <Right> "focus $page1.access.b5"
  bind $page1.access.b5 <Right> "focus $page1.access.b6"
  bind $page1.access.b7 <Right> "focus $page1.access.b8"
  bind $page1.access.b8 <Right> "focus $page1.access.b9"
  bind $page1.access.b1 <Down>  "focus $page1.access.b4"
  bind $page1.access.b2 <Down>  "focus $page1.access.b5"
  bind $page1.access.b3 <Down>  "focus $page1.access.b6"
  bind $page1.access.b4 <Down>  "focus $page1.access.b7"
  bind $page1.access.b5 <Down>  "focus $page1.access.b8"
  bind $page1.access.b6 <Down>  "focus $page1.access.b9"
  bind $page1.access.b4 <Up>    "focus $page1.access.b1"
  bind $page1.access.b5 <Up>    "focus $page1.access.b2"
  bind $page1.access.b6 <Up>    "focus $page1.access.b3"
  bind $page1.access.b7 <Up>    "focus $page1.access.b4"
  bind $page1.access.b8 <Up>    "focus $page1.access.b5"
  bind $page1.access.b9 <Up>    "focus $page1.access.b6"
  bind $page1.access.b2 <Left>  "focus $page1.access.b1"
  bind $page1.access.b3 <Left>  "focus $page1.access.b2"
  bind $page1.access.b5 <Left>  "focus $page1.access.b4"
  bind $page1.access.b6 <Left>  "focus $page1.access.b5"
  bind $page1.access.b8 <Left>  "focus $page1.access.b7"
  bind $page1.access.b9 <Left>  "focus $page1.access.b8"

  # track mouse cursor on page 2
  bind $c <1>      "set_z   $c %x %y"
  bind $c <Motion> "track_z $c %x %y"
  bind $c <Leave>  "track_z $c -1 -9999"
  
  # hints
  bind $c <Enter> \
      "status {Click to set depth coordinate}"
  bind $chip_bg_widget <Enter> \
      "status {Identifier of background chip}"
  bind $chip_bg_widget <Leave> "status {}"
  bind $chip_depth_widget <Enter> \
      "status {Depth coordinate (usually defined by pixel at floor level)}"
  bind $chip_depth_widget <Leave> "status {}"

  # help
  bind $win <F1> "chip_help $win"
  
  # deleting window is same as cancel
  wm protocol $win WM_DELETE_WINDOW "destroy $win"
  bind $win <Escape> "destroy $win"

  # swap pages
  bind $win <Control-Tab> "tabnotebook_next $win.nb; break"

  # select OK button
  focus $win.buttons.ok
}


#
# Save changes to a chip.  The function is invoked by user pressing the
# OK button in the dialog.
#
proc chip_edit_ok { win id } {
  global chip_depth_widget
  global chip_bg_widget
  global chip_access
  global chip_depth
  global chip_bg
  global chip_terrain
  global chiph
  global nchips
  global selector_widget
  global editor_widget

  # max depth coordinate
  set max [expr $chiph*4]
  
  # get depth coordinate
  set z [$chip_depth_widget get]
  if { $z < 0  ||  $z > $max  ||  ! [regexp {^[0-9][0-9]*$} $z] } {
    focus $chip_depth_widget
    tabnotebook_display $win.nb "Transparency"
    modal_dialog "Error" "Invalid depth coordinate" error
    return
  }

  # get background chip
  set bg [$chip_bg_widget get]
  if { $bg != 0 } {
    if { ![regexp {^[1-9][0-9]*$} $bg] } {
      focus $chip_bg_widget
      tabnotebook_display $win.nb "Transparency"
      modal_dialog "Error" "Invalid chip identifier" error
      return
    }
    if { $bg < 1  ||  $bg > $nchips } {
      focus $chip_bg_widget
      tabnotebook_display $win.nb "Transparency"
      modal_dialog "Error" "Chip identifier out of range" error
      return
    }
    if { $bg == $id } {
      focus $chip_bg_widget
      tabnotebook_display $win.nb "Transparency"
      modal_dialog "Error" "Chip not usable as a background" error
      return
    }
  }

  # check invalid combinations
  if { $z == 0  &&  $bg != 0 } {
    focus $chip_depth_widget
    tabnotebook_display $win.nb "Transparency"
    modal_dialog "Error" "Missing depth coordinate" error
    return
  }
  if { $bg == 0  &&  $z != 0 } {
    focus $chip_bg_widget
    tabnotebook_display $win.nb "Transparency"
    modal_dialog "Error" "Missing background chip" error
    return
  }
  
  # save depth coordinate of the chip
  set chip_depth($id) $z

  # get id of background chip
  set chip_bg($id) $bg

  # get chip accessibility
  for {set i 1} {$i < 10} {incr i} {
    set chip_terrain($id,$i) $chip_access($i)
  }

  # remake chip image
  refresh_chip $id
  refresh_view $selector_widget
  refresh_view $editor_widget

  # There are now unsaved changes present.  Note that user cannot undo
  # changes made in chip edit dialog.  However, if allowed, he could undo
  # some other changes and revert the undo status of the file at the
  # same time.  In effect, this could make the title bar show that there
  # are no unsaved changes even though some chips have been edited.  To
  # prevent this from happening, undo buffer is reset.
  set_edited 1
  reset_undo

  # close the dialog window
  destroy $win
}


#
# Track mouse cursor on z-coordinate window.  Invoked by user moving
# mouse over transparency window.
# 
proc track_z { win x y } {
  global chipw
  global chiph
  global chipzoom
  
  # remove existing level
  $win delete "curs"

  # re-insert level if mouse cursor still inside the window
  if { 0 <= $y  &&  $y < [expr $chiph*$chipzoom] } {
    # compute y-coordinate in coordinate system of the chip image
    set py [expr int($y/$chipzoom)]

    # re-insert horizontal level
    set i [expr $py*$chipzoom]
    $win create line 0 $i [expr $chipw*$chipzoom] $i -tag "curs" \
        -fill black -width 1
  } else {
    # mouse cursor leaving
    status {}
  }
}


#
# Set depth coordinate by mouse.
# 
proc set_z { win x y } {
  global chipw
  global chiph
  global chipzoom
  global chip_curz
  global chip_depth_widget
  
  # set y-coordinate in coordinate system of the chip image
  set chip_curz [expr int($y/$chipzoom)]

  # move the horizontal line on the top of the image
  show_z $win

  # update text entry
  $chip_depth_widget delete 0 end
  $chip_depth_widget insert 0 $chip_curz
}


#
# Draw horizontal line on trasparency window to mark current depth
# coordinate.
#
proc show_z { win } {
  global chipw
  global chipzoom
  global chip_curz
  
  # remove existing line
  $win delete "z"

  # paint line to canvas
  set i [expr $chip_curz*$chipzoom - 1]
  $win create line 0 $i [expr $chipw*$chipzoom] $i -tag "z" \
      -fill black -width 3
  $win create line 0 $i [expr $chipw*$chipzoom] $i -tag "z" \
      -fill red -width 1
}


#
# Show accessibility types over ith button.
#
proc chip_change_access { win i } {
  global chip_access_menu
  global chiph
  global chipw
  global chipzoom
  
  # create component for access menu
  if { $chip_access_menu == "" } {
    set chip_access_menu .chip_access
    menu $chip_access_menu -tearoff no
  }

  # fill menu
  $chip_access_menu delete 0 last
  $chip_access_menu add command -label "Grass" -underline 0 \
      -command "chip_set_access $win $i G "
  $chip_access_menu add command -label "Forest" -underline 0 \
      -command "chip_set_access $win $i F"
  $chip_access_menu add command -label "Road" -underline 0 \
      -command "chip_set_access $win $i R"
  $chip_access_menu add command -label "Water" -underline 0 \
      -command "chip_set_access $win $i W"
  $chip_access_menu add command -label "Mountain" -underline 0 \
      -command "chip_set_access $win $i M"
  $chip_access_menu add command -label "Unreachable" -underline 0 \
      -command "chip_set_access $win $i U"

  # compute location of button in screen coordinates
  set x [expr [winfo rootx $win] + $chipw*$chipzoom/6]
  set y [expr [winfo rooty $win] + $chiph*$chipzoom/6]
  
  # popup menu
  $win configure -relief sunken
  tk_popup $chip_access_menu $x $y
  $win configure -relief raised
}


#
# Swap accessibility between grass and unaccessible.
#
proc chip_swap_access { win i } {
  global chip_access
  if { $chip_access($i) != "U" } {
    chip_set_access $win $i U
  } else {
    chip_set_access $win $i G
  }
}


#
# Rotate access mode of region.
#
proc chip_rotate_access { win i } {
  global chip_access
  if { $chip_access($i) == "G" } {
    chip_set_access $win $i F
  } elseif { $chip_access($i) == "F" } {
    chip_set_access $win $i R
  } elseif { $chip_access($i) == "R" } {
    chip_set_access $win $i W
  } elseif { $chip_access($i) == "W" } {
    chip_set_access $win $i M
  } elseif { $chip_access($i) == "M" } {
    chip_set_access $win $i U
  } elseif { $chip_access($i) == "U" } {
    chip_set_access $win $i G
  }
}


#
# Set accessibility mode of a region.
#
proc chip_set_access { win i mode } {
  global chip_access

  # select button
  focus $win

  # save access mode
  set chip_access($i) $mode
 
  # change visual cue of button
  chip_show_access $win $i

  # tell about the change
  access_hint $win $i
}


#
# Show letter of accessibility over chip i.
#
proc chip_show_access { win i } {
  global chip_access
  global access_font

  # position of text within canvas
  set x 5
  set y 5

  # get letter to display
  set text $chip_access($i)

  # erase existing text
  $win delete "text"

  # black rim
  $win create text [expr $x-2] $y -anchor nw -tag "text" \
      -text $text -fill black -font $access_font
  $win create text [expr $x+2] $y -anchor nw -tag "text" \
      -text $text -fill black -font $access_font
  $win create text $x [expr $y-2] -anchor nw -tag "text" \
      -text $text -fill black -font $access_font
  $win create text $x [expr $y+2] -anchor nw -tag "text" \
      -text $text -fill black -font $access_font
  $win create text [expr $x+1] [expr $y+1] -anchor nw -tag "text" \
      -text $text -fill black -font $access_font
  $win create text [expr $x-1] [expr $y+1] -anchor nw -tag "text" \
      -text $text -fill black -font $access_font
  $win create text [expr $x+1] [expr $y-1] -anchor nw -tag "text" \
      -text $text -fill black -font $access_font
  $win create text [expr $x-1] [expr $y-1] -anchor nw -tag "text" \
      -text $text -fill black -font $access_font

  # yellow text
  $win create text $x $y -anchor nw -tag "text" -text $text \
      -fill "#ff5555" -font $access_font
}


#
# Show access mode at status bar when moving mouse pointer over a button.
#
proc access_hint { win i } {
  global chip_access
  set mode $chip_access($i)
  
  if { $mode == "G" } {
    status "Grass - player can walk here"
  } elseif { $mode == "F" } {
    status "Forest - player can walk here"
  } elseif { $mode == "R" } {
    status "Road - player can move quickly here"
  } elseif { $mode == "W" } {
    status "Water - player can sail here"
  } elseif { $mode == "M" } {
    status "Mountain - player cannot enter this area by foot"
  } elseif { $mode == "U" } {
    status "Unreachable - player cannot enter this area by any means"
  } else {
    error "Invalid mode: $mode"
  }
  focus $win
}

#
# Launch help browser from edit chip dialog.
#
proc chip_help { win } {
  # find out name of current page
  set curpage [tabnotebook_current $win.nb]

  if { $curpage == "Accessibility" } {
    help_window 5
  } elseif { $curpage == "Transparency" } {
    help_window 6
  } else {
    error "No help available for: $curpage"
  }
}