#
# map.tcl - edit and create map
# 
# Copyright (C) 2005  Toni Ronkko
#
# Peach Gardens is free software; you can redistribute it and/or modify 
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Peach Gardens is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# Nov 2004, Toni Ronkko <toni at users.playstation2-linux.com>
#
# $Id: map.tcl,v 1.3 2005/01/16 23:34:23 tronkko Exp $
#
# $Log: map.tcl,v $
# Revision 1.3  2005/01/16 23:34:23  tronkko
# default name of map
#
# Revision 1.2  2005/01/09 20:54:02  tronkko
# use select_workspace instead of refresh to ensure that all chips are
# re-drawn
#
# Revision 1.1  2004/12/09 22:16:19  tronkko
# Initial revision
#
#
#

# current map (for highlighting menu entries)
set cur_map 0


#
# Show map creation dialog, and create new map.  The optional argument
# is used as a default name for the new map.
#
proc create_map_dialog { {suggest ""} } {
  global editor_widget
  global chipset_present
  global workspace_dialog_size
  
  # Refuse to create map unless there is a chipset.  This is needed as
  # disabling menus does not disable keyboard shortcuts that we have
  # added.
  if { ! $chipset_present } {
    return
  }
  
  # offer a medium sized workspace by default
  set workspace_dialog_size 1
  
  # create workspace
  set i [create_workspace_dialog "Create Map..." map $suggest]
  if { $i >= 0 } {
    # file has now been edited
    set_edited 1
    
    # switch to the newly created map
    select_map $i
    refresh_menu
  }
}


#
# Edit name or resize map.
# 
proc edit_map_dialog {} {
  global cur_workspace
  global editor_widget

  # find out current map
  set cur $cur_workspace($editor_widget)

  # edit current workspace
  if { $cur >= 0 } {
    if { [edit_workspace_dialog "Map Properties..." $cur] } {
      # file has now been edited
      set_edited 1

      # refresh screen
      select_map $cur
      refresh_menu
      refresh_title

      # erase undo buffer as the undo events may refer to non-existent area
      reset_undo
    }
  } else {
    modal_dialog "Error" "No maps present" error
  }
}


#
# Refresh list of maps in main menu.  The function must be called
# after inserting or deleting maps.
#
proc refresh_map_menu {} {
  global map_menu
  global nworkspaces
  global workspace_name
  global editor_widget
  global workspace_used
  global workspace_type
  global chipset_present

  # empty the menu completely
  $map_menu delete 0 last

  # re-insert maps
  set count 0
  for {set i 0} {$i < $nworkspaces} {incr i} {
    if { $workspace_used($i)  &&  $workspace_type($i) == "map" } {
      $map_menu add radiobutton -underline 0 \
          -label $workspace_name($i) -variable cur_map \
          -value $i -command "select_map $i"
      incr count
    }
  }

  # allow editing maps?
  # allow editing?
  if { $chipset_present } {
    set new_state normal
    if { $count != 0 } {
      set edit_state normal
    } else {
      set edit_state disabled
    }
  } else {
    # no chipset open
    set new_state disabled
    set edit_state disabled
  }

  # re-insert actions
  $map_menu add separator
  $map_menu add command -label "New..." -state $new_state \
      -command create_map_dialog
  $map_menu add command -label "Properties..." -state $edit_state \
      -command edit_map_dialog
  $map_menu add command -label "Delete..." -state $edit_state \
      -command delete_map
}


#
# Select ith map to editor widget on the right.  The function is
# typically invoked when the user selectes a map from the menu.
#
proc select_map { i } {
  global editor_widget
  global cur_map

  # highlight maps in menu
  set cur_map $i

  # show map in widget
  select_workspace $editor_widget $i

  # pass the map name to window title
  refresh_title
}


#
# Delete current map.
#
proc delete_map {} {
  global workspace_name
  global cur_workspace
  global editor_widget

  # get index of current map
  set cur $cur_workspace($editor_widget)
  if { $cur >= 0 } {
    # confirm delete
    set name $workspace_name($cur)
    set ok [modal_dialog \
        "Delete Map?" \
        "The operation cannot be undone.  Really delete $name?" \
        question {Yes No} 0 ]
    if { $ok == 0 } {
      # erase workspace
      set map [delete_workspace $cur]

      # file has now been edited
      set_edited 1

      # reset undo buffers so that deleted buffer cannot be edited
      reset_undo

      # select some other map
      select_map $map
      refresh_menu
    } else {
      status "Not deleted"
    }
  } else {
    modal_dialog "Error" "No maps present" error
  }
}
